-------------------------------------------------------------------------------
-- (C) Altran Praxis Limited
-------------------------------------------------------------------------------
--
-- The SPARK toolset is free software; you can redistribute it and/or modify it
-- under terms of the GNU General Public License as published by the Free
-- Software Foundation; either version 3, or (at your option) any later
-- version. The SPARK toolset is distributed in the hope that it will be
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
-- Public License for more details. You should have received a copy of the GNU
-- General Public License distributed with the SPARK toolset; see file
-- COPYING3. If not, go to http://www.gnu.org/licenses for a complete copy of
-- the license.
--
--=============================================================================

separate (Dictionary)
procedure LookupScope
  (Name         : in     LexTokenManager.Lex_String;
   StopAt       : in     LexTokenManager.Lex_String;
   Scope        : in     Scopes;
   CallingScope : in     Scopes;
   Context      : in     Contexts;
   Item         :    out Symbol;
   IsVisible    :    out Boolean)
-- The extra CallingScope parameter is only used when LookUpScope is called from
-- LookUpSelectedItem because in this case only the scope we start the search in
-- is not the same as the scope where the search started (former is the visible
-- scope of the prefix package).  The extra parameter only affects the selection of
-- the correct implicit proof function that corresponds to an Ada function
is

   Region  : Symbol;
   TheItem : Symbol;

   --------------------------------------------------------------------------------

   function LookupDeclarations
     (Name    : LexTokenManager.Lex_String;
      StopAt  : LexTokenManager.Lex_String;
      Context : Contexts;
      Head    : Symbol)
     return    Symbol
   --# global in Dict;
   --#        in LexTokenManager.State;
   is
      Declaration, Item, DeclarativeItem : Symbol;
      DeclarativeItemName                : LexTokenManager.Lex_String;

      --------------------------------------------------------------------------------

      function LookupDeclaration
        (Name                : LexTokenManager.Lex_String;
         DeclarativeItem     : Symbol;
         DeclarativeItemName : LexTokenManager.Lex_String;
         Context             : Contexts;
         Declaration         : Symbol)
        return                Symbol
      --# global in Dict;
      --#        in LexTokenManager.State;
      is
         Item : Symbol;

         --------------------------------------------------------------------------------

         function LookupEnumerationLiterals (Name    : LexTokenManager.Lex_String;
                                             TheType : Symbol) return Symbol
         --# global in Dict;
         --#        in LexTokenManager.State;
         is
            Literal : Symbol;
         begin
            Trace_Lex_Str (Msg => "      In LookupEnumerationLiterals, seeking ",
                           L   => Name);

            Literal := RawDict.GetTypeFirstEnumerationLiteral (TheType);

            loop
               exit when Literal = NullSymbol;
               exit when LexTokenManager.Lex_String_Case_Insensitive_Compare
                 (Lex_Str1 => RawDict.GetEnumerationLiteralName (Literal),
                  Lex_Str2 => Name) =
                 LexTokenManager.Str_Eq;
               Literal := RawDict.GetNextEnumerationLiteral (Literal);
            end loop;

            return Literal;

         end LookupEnumerationLiterals;

         --------------------------------------------------------------------------------

      begin
         if Context = ProgramContext and then RawDict.GetDeclarationContext (Declaration) = ProofContext then
            Item := NullSymbol;
         else
            if LexTokenManager.Lex_String_Case_Insensitive_Compare (Lex_Str1 => DeclarativeItemName,
                                                                    Lex_Str2 => Name) =
              LexTokenManager.Str_Eq then
               Item := DeclarativeItem;
            elsif IsType (DeclarativeItem) and then TypeIsEnumeration (DeclarativeItem) then
               Item := LookupEnumerationLiterals (Name, DeclarativeItem);
            else
               Item := NullSymbol;
            end if;

         end if;

         return Item;

      end LookupDeclaration;

      --------------------------------------------------------------------------------

   begin
      Trace_Lex_Str (Msg => "In LookupDeclarations, seeking ",
                     L   => Name);
      Item        := NullSymbol;
      Declaration := Head;

      loop
         exit when Declaration = NullSymbol;
         DeclarativeItem := RawDict.GetDeclarationItem (Declaration);
         if RawDict.GetSymbolDiscriminant (DeclarativeItem) = OperatorSymbol then
            DeclarativeItemName := LexTokenManager.Null_String;
         else
            DeclarativeItemName := GetSimpleName (DeclarativeItem);
         end if;
         Trace_Lex_Str (Msg => "DeclarativeItemName is ",
                        L   => DeclarativeItemName);
         exit when LexTokenManager.Lex_String_Case_Insensitive_Compare
           (Lex_Str1 => StopAt,
            Lex_Str2 => LexTokenManager.Null_String) /=
           LexTokenManager.Str_Eq
           and then LexTokenManager.Lex_String_Case_Insensitive_Compare
           (Lex_Str1 => StopAt,
            Lex_Str2 => DeclarativeItemName) =
           LexTokenManager.Str_Eq;
         Item := LookupDeclaration (Name, DeclarativeItem, DeclarativeItemName, Context, Declaration);
         exit when Item /= NullSymbol;
         Declaration := RawDict.GetNextDeclaration (Declaration);
      end loop;

      return Item;

   end LookupDeclarations;

   --------------------------------------------------------------------------------

   function LookupKnownDiscriminants (Name      : in LexTokenManager.Lex_String;
                                      TheRegion : in Symbol) return Symbol
   --# global in Dict;
   --#        in LexTokenManager.State;
   is
      TheDiscriminant : Symbol;
   begin
      Trace_Lex_Str (Msg => "In LookupKnownDiscriminants, seeking ",
                     L   => Name);
      if RawDict.GetTypeDiscriminant (TheRegion) = ProtectedType then
         TheDiscriminant := RawDict.GetProtectedTypeFirstDiscriminant (TheRegion);
      else
         TheDiscriminant := RawDict.GetTaskTypeFirstDiscriminant (TheRegion);
      end if;

      loop
         exit when TheDiscriminant = NullSymbol;
         exit when LexTokenManager.Lex_String_Case_Insensitive_Compare
           (Lex_Str1 => RawDict.GetDiscriminantName (TheDiscriminant),
            Lex_Str2 => Name) =
           LexTokenManager.Str_Eq;
         TheDiscriminant := RawDict.GetNextDiscriminant (TheDiscriminant);
      end loop;
      return TheDiscriminant;
   end LookupKnownDiscriminants;

   --------------------------------------------------------------------------------

   procedure LookUpGenericFormalParameters
     (Name                : in     LexTokenManager.Lex_String;
      SubprogramOrPackage : in     Symbol;
      Parameter           :    out Symbol;
      IsVisible           :    out Boolean)
   --# global in Dict;
   --#        in LexTokenManager.State;
   --# derives IsVisible,
   --#         Parameter from Dict,
   --#                        LexTokenManager.State,
   --#                        Name,
   --#                        SubprogramOrPackage;
   is
      TheParameter : Symbol;

      function GetGenericParameterName (Sym : Symbol) return LexTokenManager.Lex_String
      --# global in Dict;
      is
         Result          : LexTokenManager.Lex_String;
         TypeOrObjectSym : Symbol;

      begin -- GetGenericParameterName
         TypeOrObjectSym := RawDict.GetGenericParameterTypeOrObject (Sym);
         Result          := GetSimpleName (TypeOrObjectSym);
         return Result;
      end GetGenericParameterName;

   begin -- LookUpGenericFormalParameters
      Trace_Lex_Str (Msg => "      In LookupGenericFormalParameters, seeking ",
                     L   => Name);

      IsVisible := False;       -- default if not found
      Parameter := NullSymbol;  -- default if not found

      -- GenericTBD extend this to handle generic package formal later
      if IsPackage (SubprogramOrPackage) then
         TheParameter := RawDict.GetPackageFirstGenericParameter (SubprogramOrPackage);
      else
         TheParameter := RawDict.GetSubprogramFirstGenericParameter (SubprogramOrPackage);
      end if;

      loop
         exit when TheParameter = NullSymbol;
         if LexTokenManager.Lex_String_Case_Insensitive_Compare
           (Lex_Str1 => GetGenericParameterName (TheParameter),
            Lex_Str2 => Name) =
           LexTokenManager.Str_Eq then
            -- actually want the generic entity not the generic parameter
            Parameter := RawDict.GetGenericParameterTypeOrObject (TheParameter);
            IsVisible := True;
         end if;

         TheParameter := RawDict.GetGenericParameterNext (TheParameter);
      end loop;

   end LookUpGenericFormalParameters;

   --------------------------------------------------------------------------------

   function LookupVisibleDeclarations
     (Name      : LexTokenManager.Lex_String;
      TheRegion : Symbol;
      Context   : Contexts)
     return      Symbol
   --# global in Dict;
   --#        in LexTokenManager.State;
   is
      Item : Symbol;

      ------------------------------------------------------------------------------

      function LookupOwnVariables (Name       : LexTokenManager.Lex_String;
                                   ThePackage : Symbol) return Symbol
      --# global in Dict;
      --#        in LexTokenManager.State;
      is
         OwnVariable, Variable : Symbol;
      begin
         Trace_Lex_Str (Msg => "   In LookupOwnVariables, seeking ",
                        L   => Name);

         OwnVariable := RawDict.GetPackageOwnVariables (ThePackage);

         loop
            if OwnVariable = NullSymbol then
               Variable := NullSymbol;
               exit;
            end if;
            Variable := RawDict.GetOwnVariableVariable (OwnVariable);
            exit when LexTokenManager.Lex_String_Case_Insensitive_Compare
              (Lex_Str1 => RawDict.GetVariableName (Variable),
               Lex_Str2 => Name) =
              LexTokenManager.Str_Eq;
            OwnVariable := RawDict.GetNextOwnVariable (OwnVariable);
         end loop;

         return Variable;

      end LookupOwnVariables;

      ------------------------------------------------------------------------------

      function LookupOwnTasks (Name       : LexTokenManager.Lex_String;
                               ThePackage : Symbol) return Symbol
      --# global in Dict;
      --#        in LexTokenManager.State;
      is
         OwnTask, Variable : Symbol;
      begin
         Trace_Lex_Str (Msg => "   In LookupOwnTasks, seeking ",
                        L   => Name);

         OwnTask := RawDict.GetPackageTaskList (ThePackage);

         loop
            if OwnTask = NullSymbol then
               Variable := NullSymbol;
               exit;
            end if;
            Variable := RawDict.GetOwnTaskVariable (OwnTask);
            exit when LexTokenManager.Lex_String_Case_Insensitive_Compare
              (Lex_Str1 => RawDict.GetVariableName (Variable),
               Lex_Str2 => Name) =
              LexTokenManager.Str_Eq;
            OwnTask := RawDict.GetNextOwnTask (OwnTask);
         end loop;

         return Variable;

      end LookupOwnTasks;

      ------------------------------------------------------------------------------

      function LookUpPackageVisibleDeclarations (ThePackage : Symbol) return Symbol
      --# global in Context;
      --#        in Dict;
      --#        in LexTokenManager.State;
      --#        in Name;
      is
         Item  : Symbol;
         Dummy : Boolean;
      begin
         Trace_Lex_Str (Msg => "   In LookupPackageVisibleDeclarations, seeking ",
                        L   => Name);
         if LexTokenManager.Lex_String_Case_Insensitive_Compare (Lex_Str1 => GetSimpleName (ThePackage),
                                                                 Lex_Str2 => Name) =
           LexTokenManager.Str_Eq then
            Item := ThePackage;
         else
            Item :=
              LookupDeclarations
              (Name,
               LexTokenManager.Null_String,
               Context,
               RawDict.GetPackageFirstVisibleDeclaration (ThePackage));

            if Item = NullSymbol then

               Item :=
                 LookupDeclarations
                 (Name,
                  LexTokenManager.Null_String,
                  Context,
                  RawDict.GetPackageVisibleRenamingDeclarations (ThePackage));

               if Item = NullSymbol and then Is_Generic_Package (The_Symbol => ThePackage) then
                  --# accept Flow, 10, Dummy, "Expected ineffective assignment to Dummy";
                  LookUpGenericFormalParameters (Name, ThePackage, Item, Dummy);
                  --# end accept;
               end if;

               if Item = NullSymbol and then Context = ProofContext then
                  Item := LookupOwnVariables (Name, ThePackage);
                  if Item = NullSymbol then
                     Item := LookupOwnTasks (Name, ThePackage);
                  end if;
               end if;
            end if;
         end if;
         --# accept Flow, 33, Dummy, "Expected to be neither referenced nor exported Dummy";
         return Item;
      end LookUpPackageVisibleDeclarations;

      -----------------------------------

      function LookUpProtectedTypeVisibleDeclarations (TheProtectedType : Symbol) return Symbol
      --# global in Context;
      --#        in Dict;
      --#        in LexTokenManager.State;
      --#        in Name;
      is
         Item : Symbol;
      begin
         Trace_Lex_Str (Msg => "   In LookupProtectedTypeVisibleDeclarations, seeking ",
                        L   => Name);
         if LexTokenManager.Lex_String_Case_Insensitive_Compare (Lex_Str1 => GetSimpleName (TheProtectedType),
                                                                 Lex_Str2 => Name) =
           LexTokenManager.Str_Eq then
            -- if we find PT (say) when looking from within PT then what we want to return is the
            -- implicitly declared abstract own variable of PT rather than the type itself
            Item := RawDict.GetOwnVariableVariable (RawDict.GetProtectedTypeOwnVariable (TheProtectedType));
         else
            Item :=
              LookupDeclarations
              (Name,
               LexTokenManager.Null_String,
               Context,
               RawDict.GetProtectedTypeFirstVisibleDeclaration (TheProtectedType));
         end if;

         -- If not found then check the known discriminants
         if Item = NullSymbol then
            Item := LookupKnownDiscriminants (Name, TheProtectedType);
         end if;

         --   -- if we don't find what we are looking for in the protected type we need to look again in
         --   -- the visible declarations of the enclosing package since these are also directly visible
         --   if Item = NullSymbol then
         --      Item := LookUpPackageVisibleDeclarations (GetEnclosingPackage (VisibleScope (TheProtectedType)));
         --   end if;

         return Item;
      end LookUpProtectedTypeVisibleDeclarations;

      -----------------------------------

      function LookUpTaskTypeVisibleDeclarations (TheTaskType : Symbol) return Symbol
      --# global in Dict;
      --#        in LexTokenManager.State;
      --#        in Name;
      is
         Item : Symbol;
      begin
         Trace_Lex_Str (Msg => "   In LookupTaskTypeVisibleDeclarations, seeking ",
                        L   => Name);
         -- check the know discriminants
         Item := LookupKnownDiscriminants (Name, TheTaskType);

         -- if we don't find what we are looking for in the task type we need to look again in
         -- the visible declarations of the enclosing package since these are also directly visible
         -- if Item = NullSymbol then
         --   Item := LookUpPackageVisibleDeclarations (GetEnclosingPackage (VisibleScope (TheTaskType)));
         -- end if;

         return Item;
      end LookUpTaskTypeVisibleDeclarations;

      -----------------------------------

   begin --LookupVisibleDeclarations
      Trace_Lex_Str (Msg => "In LookupVisibleDeclarations, seeking ",
                     L   => Name);

      if IsPackage (TheRegion) then
         Item := LookUpPackageVisibleDeclarations (TheRegion);
      elsif IsType (TheRegion) and then TypeIsProtected (TheRegion) then
         Item := LookUpProtectedTypeVisibleDeclarations (TheRegion);
      elsif IsType (TheRegion) and then TypeIsTask (TheRegion) then
         Item := LookUpTaskTypeVisibleDeclarations (TheRegion);
      else -- error, nothing else has visible scope
         SystemErrors.Fatal_Error (Sys_Err => SystemErrors.Invalid_Symbol_Table,
                                   Msg     => "Invalid LookupVisibleDeclarations");
         Item := NullSymbol; -- can't be reached, there just to avoid DFerr
      end if;

      return Item;

   end LookupVisibleDeclarations;

   ------------------------------------------------------------------------------

   procedure LookupLocalDeclarations
     (Name      : in     LexTokenManager.Lex_String;
      StopAt    : in     LexTokenManager.Lex_String;
      Region    : in     Symbol;
      Context   : in     Contexts;
      ScopeType : in     Visibility;
      Item      :    out Symbol;
      IsVisible :    out Boolean)
   --# global in CallingScope;
   --#        in Dict;
   --#        in LexTokenManager.State;
   --# derives IsVisible,
   --#         Item      from CallingScope,
   --#                        Context,
   --#                        Dict,
   --#                        LexTokenManager.State,
   --#                        Name,
   --#                        Region,
   --#                        ScopeType,
   --#                        StopAt;
   is

      TheItem         : Symbol;
      TheDiscriminant : SymbolDiscriminant;

      ------------------------------------------------------------------------------

      function LookupPackageBody
        (Name       : LexTokenManager.Lex_String;
         StopAt     : LexTokenManager.Lex_String;
         ThePackage : Symbol;
         Context    : Contexts;
         ScopeType  : Visibility)
        return       Symbol
      --# global in Dict;
      --#        in LexTokenManager.State;
      is
         Item : Symbol;

         ------------------------------------------------------------------------------

         function LookupRefinementConstituents (Name       : LexTokenManager.Lex_String;
                                                ThePackage : Symbol) return Symbol
         --# global in Dict;
         --#        in LexTokenManager.State;
         is
            AbstractOwnVariables : Iterator;
            Constituent          : Symbol;

            ------------------------------------------------------------------------------

            function LookupConstituents (Name    : LexTokenManager.Lex_String;
                                         Subject : Symbol) return Symbol
            --# global in Dict;
            --#        in LexTokenManager.State;
            is
               Constituents : Iterator;
               Constituent  : Symbol;
            begin
               Trace_Lex_Str (Msg => "         In LookupConstituents, seeking ",
                              L   => Name);

               Constituents := FirstConstituent (Subject);

               loop
                  if IsNullIterator (Constituents) then
                     Constituent := NullSymbol;
                     exit;
                  end if;
                  Constituent := CurrentSymbol (Constituents);
                  exit when LexTokenManager.Lex_String_Case_Insensitive_Compare
                    (Lex_Str1 => GetSimpleName (Constituent),
                     Lex_Str2 => Name) =
                    LexTokenManager.Str_Eq
                    and then not IsOwnVariable (Constituent);
                  Constituents := NextSymbol (Constituents);
               end loop;

               return Constituent;

            end LookupConstituents;

            ------------------------------------------------------------------------------

         begin
            Trace_Lex_Str (Msg => "      In LookupRefinementConstituents, seeking ",
                           L   => Name);

            AbstractOwnVariables := FirstAbstractOwnVariable (ThePackage);
            Constituent          := NullSymbol;

            loop
               exit when IsNullIterator (AbstractOwnVariables);
               Constituent := LookupConstituents (Name, CurrentSymbol (AbstractOwnVariables));
               exit when Constituent /= NullSymbol;
               AbstractOwnVariables := NextSymbol (AbstractOwnVariables);
            end loop;

            return Constituent;

         end LookupRefinementConstituents;

         ------------------------------------------------------------------------------

      begin
         Trace_Lex_Str (Msg => "In LookupPackageBody, seeking ",
                        L   => Name);
         if ScopeType = Privat then
            Item := LookupDeclarations (Name, StopAt, Context, RawDict.GetPackageFirstPrivateDeclaration (ThePackage));

         else
            Item := LookupDeclarations (Name, StopAt, Context, RawDict.GetPackageFirstLocalDeclaration (ThePackage));

            if Item = NullSymbol then
               Item := LookupDeclarations (Name, StopAt, Context, RawDict.GetPackageLocalRenamingDeclarations (ThePackage));
            end if;

            if Item = NullSymbol and then Context = ProofContext then
               Item := LookupRefinementConstituents (Name, ThePackage);
            end if;

            if Item = NullSymbol then
               Item := LookupDeclarations (Name, StopAt, Context, RawDict.GetPackageFirstPrivateDeclaration (ThePackage));
            end if;
         end if;

         if Item = NullSymbol then
            Item := LookupVisibleDeclarations (Name, ThePackage, Context);
         end if;

         return Item;

      end LookupPackageBody;

      ------------------------------------------------------------------------------

      procedure LookupSubprogramBody
        (Name       : in     LexTokenManager.Lex_String;
         StopAt     : in     LexTokenManager.Lex_String;
         Subprogram : in     Symbol;
         Context    : in     Contexts;
         Item       :    out Symbol;
         IsVisible  :    out Boolean)
      --# global in Dict;
      --#        in LexTokenManager.State;
      --# derives IsVisible,
      --#         Item      from Context,
      --#                        Dict,
      --#                        LexTokenManager.State,
      --#                        Name,
      --#                        StopAt,
      --#                        Subprogram;
      is

         TheItem : Symbol;

         ------------------------------------------------------------------------------

         procedure LookupSubprogramParameters
           (Name       : in     LexTokenManager.Lex_String;
            Subprogram : in     Symbol;
            Context    : in     Contexts;
            Parameter  :    out Symbol;
            IsVisible  :    out Boolean)
         --# global in Dict;
         --#        in LexTokenManager.State;
         --# derives IsVisible from Context,
         --#                        Dict,
         --#                        LexTokenManager.State,
         --#                        Name,
         --#                        Subprogram &
         --#         Parameter from Dict,
         --#                        LexTokenManager.State,
         --#                        Name,
         --#                        Subprogram;
         is

            TheParameter : Symbol;

            ------------------------------------------------------------------------------

            procedure LookupGlobalVariables
              (Name       : in     LexTokenManager.Lex_String;
               Subprogram : in     Symbol;
               Context    : in     Contexts;
               Variable   :    out Symbol;
               IsVisible  :    out Boolean)
            --# global in Dict;
            --#        in LexTokenManager.State;
            --# derives IsVisible from Context,
            --#                        Dict,
            --#                        LexTokenManager.State,
            --#                        Name,
            --#                        Subprogram &
            --#         Variable  from Dict,
            --#                        LexTokenManager.State,
            --#                        Name,
            --#                        Subprogram;
            is
               GlobalVariable, TheVariable : Symbol;
            begin
               Trace_Lex_Str (Msg => "         In LookupGlobalVariables, seeking ",
                              L   => Name);

               GlobalVariable :=
                 RawDict.GetSubprogramFirstGlobalVariable (GetAbstraction (Subprogram, LocalScope (Subprogram)), Subprogram);

               loop
                  if GlobalVariable = NullSymbol then
                     Variable  := NullSymbol;
                     IsVisible := False;
                     exit;
                  end if;
                  TheVariable := RawDict.GetGlobalVariableVariable (GlobalVariable);
                  if LexTokenManager.Lex_String_Case_Insensitive_Compare
                    (Lex_Str1 => GetSimpleName (TheVariable),
                     Lex_Str2 => Name) =
                    LexTokenManager.Str_Eq
                    and then not RawDict.GetGlobalVariablePrefixNeeded (GlobalVariable) then
                     Variable  := TheVariable;
                     IsVisible := Context = ProofContext or else IsDeclared (TheVariable);
                     exit;
                  end if;
                  GlobalVariable := RawDict.GetNextGlobalVariable (GlobalVariable);
               end loop;

            end LookupGlobalVariables;

            ------------------------------------------------------------------------------

         begin
            Trace_Lex_Str (Msg => "      In LookupSubprogramParameters, seeking ",
                           L   => Name);

            TheParameter := RawDict.GetSubprogramFirstParameter (Subprogram);

            loop
               exit when TheParameter = NullSymbol;
               exit when LexTokenManager.Lex_String_Case_Insensitive_Compare
                 (Lex_Str1 => RawDict.GetSubprogramParameterName (TheParameter),
                  Lex_Str2 => Name) =
                 LexTokenManager.Str_Eq;
               TheParameter := RawDict.GetNextSubprogramParameter (TheParameter);
            end loop;

            if TheParameter = NullSymbol then
               LookupGlobalVariables (Name, Subprogram, Context, Parameter, IsVisible);
            else
               Parameter := TheParameter;
               IsVisible := True;
            end if;

         end LookupSubprogramParameters;

         ------------------------------------------------------------------------------

      begin
         Trace_Lex_Str (Msg => "In LookupSubprogramBody, seeking ",
                        L   => Name);
         TheItem := LookupDeclarations (Name, StopAt, Context, RawDict.GetSubprogramFirstDeclaration (Subprogram));
         if TheItem = NullSymbol then
            TheItem := LookupDeclarations (Name, StopAt, Context, RawDict.GetSubprogramRenamingDeclarations (Subprogram));
         end if;

         if TheItem = NullSymbol then
            LookupSubprogramParameters (Name, Subprogram, Context, TheItem, IsVisible);
            if TheItem = NullSymbol and then Is_Generic_Subprogram (The_Symbol => Subprogram) then
               LookUpGenericFormalParameters (Name, Subprogram, Item, IsVisible);
            else
               Item := TheItem;
            end if;
         else
            Item      := TheItem;
            IsVisible := True;
         end if;

      end LookupSubprogramBody;

      ------------------------------------------------------------------------------

      function LookupLoop (Name    : LexTokenManager.Lex_String;
                           TheLoop : Symbol) return Symbol
      --# global in Dict;
      --#        in LexTokenManager.State;
      is
         Item : Symbol;

         -----------------------------------------------------------------------------

         function LookupLoopParameter (Name    : LexTokenManager.Lex_String;
                                       ForLoop : Symbol) return Symbol
         --# global in Dict;
         --#        in LexTokenManager.State;
         is
            Item : Symbol;
         begin
            Trace_Lex_Str (Msg => "   In LookupLoopParameter, seeking ",
                           L   => Name);

            if LexTokenManager.Lex_String_Case_Insensitive_Compare
              (Lex_Str1 => GetSimpleName (GetLoopParameter (ForLoop)),
               Lex_Str2 => Name) =
              LexTokenManager.Str_Eq then
               Item := GetLoopParameter (ForLoop);
            else
               Item := NullSymbol;
            end if;

            return Item;

         end LookupLoopParameter;

         ------------------------------------------------------------------------------

      begin
         Trace_Lex_Str (Msg => "In LookupLoop, seeking ",
                        L   => Name);

         if IsForLoop (TheLoop) then
            Item := LookupLoopParameter (Name, TheLoop);
         else
            Item := NullSymbol;
         end if;

         return Item;

      end LookupLoop;

      --------------------------------------------------------------------------------

      function LookupType
        (Name    : LexTokenManager.Lex_String;
         Context : Contexts;
         TheType : Symbol)
        return    Symbol
      --# global in Dict;
      --#        in LexTokenManager.State;
      --#        in ScopeType;
      is
         Item : Symbol;

         --------------------------------------------------------------------------------

         function LookupRecordComponents (Name    : LexTokenManager.Lex_String;
                                          TheType : Symbol) return Symbol
         --# global in Dict;
         --#        in LexTokenManager.State;
         is
            Component      : Symbol;
            CurrentRecord  : Symbol;
            FirstComponent : Symbol;

         begin
            Trace_Lex_Str (Msg => "      In LookupRecordComponents, seeking ",
                           L   => Name);
            CurrentRecord := TheType;
            loop
               FirstComponent := RawDict.GetTypeFirstRecordComponent (CurrentRecord);
               Component      := FirstComponent;

               loop
                  exit when Component = NullSymbol; -- did not find

                  exit when LexTokenManager.Lex_String_Case_Insensitive_Compare
                    (Lex_Str1 => RawDict.GetRecordComponentName (Component),
                     Lex_Str2 => Name) =
                    LexTokenManager.Str_Eq; -- found
                  Component := RawDict.GetNextRecordComponent (Component);
               end loop;
               exit when Component /= NullSymbol; -- carry success out of outer loop

               -- if we get here we failed to find field in local declarations
               -- so we search in inherited fields if there are any
               exit when not RecordComponentIsInherited (FirstComponent); -- no inherited fields

               -- restart search in inherited fields
               CurrentRecord := RawDict.GetRecordComponentType (FirstComponent);
            end loop;

            return Component;

         end LookupRecordComponents;

         --------------------------------------------------------------------------------

         function LookupProtectedElements
           (Name    : LexTokenManager.Lex_String;
            Context : Contexts;
            TheType : Symbol)
           return    Symbol
         --# global in Dict;
         --#        in LexTokenManager.State;
         is
            Item : Symbol;
            It   : Iterator;

            function LookupRefinementConstituents (Name    : LexTokenManager.Lex_String;
                                                   TheType : Symbol) return Symbol
            --# global in Dict;
            --#        in LexTokenManager.State;
            is
               Constituents : Iterator;
               Constituent  : Symbol;
            begin
               Constituents := FirstConstituent (GetProtectedTypeOwnVariable (TheType));
               loop
                  if IsNullIterator (Constituents) then
                     Constituent := NullSymbol;
                     exit;
                  end if;
                  Constituent := CurrentSymbol (Constituents);
                  exit when LexTokenManager.Lex_String_Case_Insensitive_Compare
                    (Lex_Str1 => GetSimpleName (Constituent),
                     Lex_Str2 => Name) =
                    LexTokenManager.Str_Eq;
                  Constituents := NextSymbol (Constituents);
               end loop;

               return Constituent;
            end LookupRefinementConstituents;

         begin
            Trace_Lex_Str (Msg => "      In LookupProtectedElements, seeking ",
                           L   => Name);
            Item := NullSymbol;
            It   := FirstProtectedElement (TheType);
            while not IsNullIterator (It) loop
               if LexTokenManager.Lex_String_Case_Insensitive_Compare
                 (Lex_Str1 => GetSimpleName (CurrentSymbol (It)),
                  Lex_Str2 => Name) =
                 LexTokenManager.Str_Eq then
                  Item := CurrentSymbol (It);
                  exit;
               end if;
               It := NextSymbol (It);
            end loop;

            -- As we add each protected element to the Dictionary they first get added as
            -- refinement cosntiuents (in proof context) and then as variables.  The variable
            -- add must match each up with the refinement constituent already added so we
            -- need a special look up here to find them
            if Item = NullSymbol and then Context = ProofContext then
               Item := LookupRefinementConstituents (Name, TheType);
            end if;

            -- if Item is null at this point then we have not found a match in the private part
            -- of the private type but we must also look in the visible part since things here
            -- are also directly visible
            if Item = NullSymbol then
               Item := LookupVisibleDeclarations (Name, TheType, Context);
            end if;
            return Item;
         end LookupProtectedElements;

         --------------------------------------------------------------------------------

         function LookupProtectedLocalScope
           (Name    : LexTokenManager.Lex_String;
            Context : Contexts;
            TheType : Symbol)
           return    Symbol
         --# global in Dict;
         --#        in LexTokenManager.State;
         --#        in ScopeType;
         is
            Item : Symbol;
         begin
            Trace_Lex_Str (Msg => "      In LookupProtectedLocalScope, seeking ",
                           L   => Name);
            -- Search local declarations if we are in local scope.  If this fails (or if we
            -- are in private scope), then search protected elements etc., by calling
            -- LookupProtectedElements
            Item := NullSymbol;
            if ScopeType = Local then
               -- search starts in protected body where we seek local declarations
               Item :=
                 LookupDeclarations
                 (Name,
                  LexTokenManager.Null_String,
                  Context,
                  RawDict.GetProtectedTypeFirstLocalDeclaration (TheType));

               if Item = NullSymbol then
                  -- we need to look at the part of the enclosing package body which is "above" the
                  -- protected type declaration
                  Item :=
                    LookupPackageBody
                    (Name,
                     GetSimpleName (TheType), -- don't look below the type declaration
                     GetEnclosingPackage (LocalScope (TheType)),
                     Context,
                     Local);
               end if;
            end if;

            if Item = NullSymbol then
               -- we either failed to find the item in the protected body or we skipped
               -- it completely because ScopeType = Privat on entry; so now we search
               -- the private part of the protected type
               Item := LookupProtectedElements (Name, Context, TheType);
            end if;
            return Item;
         end LookupProtectedLocalScope;

         --------------------------------------------------------------------------------

         function LookupTaskLocalScope
           (Name    : LexTokenManager.Lex_String;
            Context : Contexts;
            TheType : Symbol)
           return    Symbol
         --# global in Dict;
         --#        in LexTokenManager.State;
         is
            Item : Symbol;

            function LookupGlobalVariables
              (Name    : LexTokenManager.Lex_String;
               TheType : Symbol;
               Context : Contexts)
              return    Symbol
            --# global in Dict;
            --#        in LexTokenManager.State;
            is
               GlobalVariable, TheVariable, Result : Symbol;
            begin
               Trace_Lex_Str (Msg => "            In LookupGlobalVariables, seeking ",
                              L   => Name);
               GlobalVariable :=
                 RawDict.GetTaskTypeFirstGlobalVariable (GetAbstraction (TheType, LocalScope (TheType)), TheType);

               loop
                  if GlobalVariable = NullSymbol then
                     Result := NullSymbol;
                     exit;
                  end if;
                  TheVariable := RawDict.GetGlobalVariableVariable (GlobalVariable);
                  if LexTokenManager.Lex_String_Case_Insensitive_Compare
                    (Lex_Str1 => GetSimpleName (TheVariable),
                     Lex_Str2 => Name) =
                    LexTokenManager.Str_Eq
                    and then not RawDict.GetGlobalVariablePrefixNeeded (GlobalVariable) then
                     Result := TheVariable;
                     if Context = ProgramContext and then not IsDeclared (TheVariable) then
                        Result := NullSymbol;
                     end if;
                     exit;
                  end if;
                  GlobalVariable := RawDict.GetNextGlobalVariable (GlobalVariable);
               end loop;
               return Result;
            end LookupGlobalVariables;

         begin
            -- search starts in task body where we seek local declarations
            Trace_Lex_Str (Msg => "         In LookupTaskLocalScope, seeking ",
                           L   => Name);
            Item :=
              LookupDeclarations (Name, LexTokenManager.Null_String, Context, RawDict.GetTaskTypeFirstLocalDeclaration (TheType));

            if Item = NullSymbol then
               -- look up globals
               Item := LookupGlobalVariables (Name, TheType, Context);
            end if;

            --pna-- Removed this clause because it is not the role of LookUpScope to step out into the
            --pna-- enclosing package; that should be done by LookUpItem if LookUpScope fails
            --pna--
            --pna-- if Item = NullSymbol then
            --pna--    -- we need to look at the part of the enclosing package body which is "above" the
            --pna--    -- task type declaration
            --pna--    Item := LookupPackageBody (Name,
            --pna--                               GetSimpleName (TheType), -- don't look below the type declaration
            --pna--                               GetEnclosingPackage (LocalScope (TheType)),
            --pna--                               Context,
            --pna--                               Local);
            --pna-- end if;

            if Item = NullSymbol then
               Item := LookupKnownDiscriminants (Name, TheType);
            end if;

            return Item;
         end LookupTaskLocalScope;

         --------------------------------------------------------------------------------

      begin
         Trace_Lex_Str (Msg => "      In LookupType, seeking ",
                        L   => Name);
         case RawDict.GetTypeDiscriminant (TheType) is
            when RecordType =>
               Item := LookupRecordComponents (Name, TheType);
            when ProtectedType =>
               Item := LookupProtectedLocalScope (Name, Context, TheType);
            when TaskType =>
               Item := LookupTaskLocalScope (Name, Context, TheType);
               --Item := LookupVisibleDeclarations (Name, TheType, Context);
            when others =>
               Item := NullSymbol;
         end case;

         return Item;

      end LookupType;

      ------------------------------------------------------------------------------

      --      function LookupSubcomponents (Name         : LexTokenManager.Lex_String;
      --                                    Subcomponent : Symbol) return Symbol
      --         --# global Dict;
      --     is
      --         Current           : Symbol;
      --         CurrentRecord     : Symbol;
      --         FirstSubComponent : Symbol;
      --      begin
      --         CurrentRecord := Subcomponent;
      --         loop
      --            FirstSubComponent := CurrentRecord;
      --            Current := FirstSubComponent;
      --            loop
      --               exit when Current = NullSymbol;
      --               exit when RawDict.GetRecordComponentName
      --                 (RawDict.GetSubcomponentComponent (Current)) = Name;
      --               Current := RawDict.GetNextSubcomponent (Current);
      --            end loop;
      --            exit when Current /= NullSymbol; -- carry success out of outer loop
      --
      --            -- if we get here then we failed to find the desired field at the first attempt
      --            -- so we move to the first inherited field (if there is one) and try again
      --            exit when not RecordComponentIsInherited (RawDict.GetSubcomponentComponent (FirstSubComponent));
      --            CurrentRecord := RawDict.GetSubcomponentSubcomponents (FirstSubComponent);
      --         end loop;
      --         return Current;
      --
      --      end LookupSubcomponents;

      ------------------------------------------------------------------------------

      function LookupSubcomponents
        (Name         : LexTokenManager.Lex_String;
         Subcomponent : Symbol;
         CallingScope : Scopes)
        return         Symbol
      --# global in Dict;
      --#        in LexTokenManager.State;
      is
         Current           : Symbol;
         CurrentRecord     : Symbol;
         FirstSubComponent : Symbol;
         Found             : Boolean := False;
         IsPrivateField    : Boolean := False;
      begin
         Trace_Lex_Str (Msg => "      In LookupSubcomponents, seeking ",
                        L   => Name);
         CurrentRecord := Subcomponent;
         loop
            FirstSubComponent := CurrentRecord;
            Current           := FirstSubComponent;
            if not IsPrivateField then
               loop -- search fields at current depth of current record
                  exit when Current = NullSymbol;
                  Found :=
                    LexTokenManager.Lex_String_Case_Insensitive_Compare
                    (Lex_Str1 => RawDict.GetRecordComponentName (RawDict.GetSubcomponentComponent (Current)),
                     Lex_Str2 => Name) =
                    LexTokenManager.Str_Eq;
                  exit when Found;
                  Current := RawDict.GetNextSubcomponent (Current);
               end loop;
            end if;
            exit when Found; -- carry success out of outer loop

            -- if we get here then we failed to find the desired field at the current depth
            -- so we move to the first inherited field (if there is one and if it not a private
            -- extension) and try again
            exit when not RecordComponentIsInherited (RawDict.GetSubcomponentComponent (FirstSubComponent));
            -- no more inherited fields
            IsPrivateField := IsPrivateType (GetType (CurrentRecord), CallingScope);
            CurrentRecord  := RawDict.GetSubcomponentSubcomponents (FirstSubComponent);
         end loop;
         if not Found then
            Current := NullSymbol;
         end if;
         return Current;

      end LookupSubcomponents;

      ------------------------------------------------------------------------------

   begin  --LookupLocalDeclarations

      --  first part of if traps things like declaring a variable P in
      --  a procedure P and also handles implicit return variables in
      --  function proof annotations.
      Trace_Lex_Str (Msg => "   In LookupLocalDeclarations, seeking ",
                     L   => Name);
      TheDiscriminant := RawDict.GetSymbolDiscriminant (Region);

      if LexTokenManager.Lex_String_Case_Insensitive_Compare (Lex_Str1 => GetSimpleName (Region),
                                                              Lex_Str2 => Name) =
        LexTokenManager.Str_Eq
        and then
        -- Trap finding of A in X.A in nested record case
        -- Note that this case also means that seeking PT from the private part or body of PT
        -- returns the type symbol not the implicit own variable of the protected type
        TheDiscriminant /= SubcomponentSymbol
        and then TheDiscriminant /= VariableSymbol
        and then TheDiscriminant /= SubprogramParameterSymbol
        and then
        -- This line added as a result of CFR 1738. It allows the rest of the procedure to be
        -- used in the sole situation where we are performing lookup on a member of a record
        -- that has the same name as its own type.
        (TheDiscriminant /= TypeSymbol or else RawDict.GetTypeDiscriminant (Region) /= RecordType) then
         Item      := Region;
         IsVisible := True;
      else
         case RawDict.GetSymbolDiscriminant (Region) is
            when PackageSymbol =>
               TheItem   := LookupPackageBody (Name, StopAt, Region, Context, ScopeType);
               Item      := TheItem;
               IsVisible := TheItem /= NullSymbol;
            when SubprogramSymbol =>
               LookupSubprogramBody (Name, StopAt, Region, Context, Item, IsVisible);
            when LoopSymbol =>
               TheItem   := LookupLoop (Name, Region);
               Item      := TheItem;
               IsVisible := TheItem /= NullSymbol;
            when ImplicitReturnVariableSymbol | QuantifiedVariableSymbol =>
               Item      := NullSymbol;
               IsVisible := False;
            when SubcomponentSymbol =>
               TheItem   := LookupSubcomponents (Name, RawDict.GetSubcomponentSubcomponents (Region), CallingScope);
               Item      := TheItem;
               IsVisible := TheItem /= NullSymbol;
            when VariableSymbol =>
               TheItem   := LookupSubcomponents (Name, RawDict.GetVariableSubComponents (Region), CallingScope);
               Item      := TheItem;
               IsVisible := TheItem /= NullSymbol;
            when SubprogramParameterSymbol =>
               TheItem   := LookupSubcomponents (Name, RawDict.GetSubprogramParameterSubComponents (Region), CallingScope);
               Item      := TheItem;
               IsVisible := TheItem /= NullSymbol;
            when others =>
               -- must be a type (record, task or protected)
               TheItem   := LookupType (Name, Context, Region);
               Item      := TheItem;
               IsVisible := TheItem /= NullSymbol;
         end case;
      end if;

   end LookupLocalDeclarations;

   ------------------------------------------------------------------------------

begin
   Region := GetRegion (Scope);

   Trace_Lex_Str (Msg => "In LookupScope, seeking ",
                  L   => Name);
   Trace_Sym (Msg   => "   in ",
              Sym   => Region,
              Scope => Scope);
   Trace_Lex_Str (Msg => "   with StopAt set to ",
                  L   => StopAt);

   case Scope.TypeOfScope is
      when Visible =>
         TheItem   := LookupVisibleDeclarations (Name, Region, Context);
         IsVisible := TheItem /= NullSymbol;
      when Local | Privat =>
         LookupLocalDeclarations (Name, StopAt, Region, Context, Scope.TypeOfScope, TheItem, IsVisible);
   end case;

   if Context = ProofContext and then IsAdaFunction (TheItem) then
      TheItem := RawDict.GetSubprogramImplicitProofFunction (GetAbstraction (TheItem, CallingScope), TheItem);
   end if;

   Item := TheItem;
   Trace_Sym (Msg   => "Found in LookUpScope ",
              Sym   => TheItem,
              Scope => Scope);
end LookupScope;
