-------------------------------------------------------------------------------
-- (C) Altran Praxis Limited
-------------------------------------------------------------------------------
--
-- The SPARK toolset is free software; you can redistribute it and/or modify it
-- under terms of the GNU General Public License as published by the Free
-- Software Foundation; either version 3, or (at your option) any later
-- version. The SPARK toolset is distributed in the hope that it will be
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
-- Public License for more details. You should have received a copy of the GNU
-- General Public License distributed with the SPARK toolset; see file
-- COPYING3. If not, go to http://www.gnu.org/licenses for a complete copy of
-- the license.
--
--=============================================================================

separate (Sem.CompUnit)
procedure ExpressionTypeFromContext
  (Exp_Node         : in     STree.SyntaxNode;
   E_Stack          : in     ExpStack.ExpStackType;
   T_Stack          : in     TypeContextStack.TStackType;
   New_Context_Type :    out Dictionary.Symbol) is
   TopOfExpStack   : Exp_Record;
   ParameterNameOK : Boolean;
   ParamSymbol     : Dictionary.Symbol;
   TNode2          : STree.SyntaxNode;
begin
   -- Assume NodeType(ExpNode) = expression or
   --        NodeType(ExpNode) = annotation_expression

   ------------------------------------------------------------------------
   -- Possible parent nodes are:
   --
   -- Group 1 - Expressions appearing in statements or declarations
   --
   -- constant_declaration
   -- variable_declaration
   -- assignment_statement
   -- condition
   -- case_statement
   -- return_statement
   -- delay_statement
   --
   -- Group 2 - Expressions appearing in expressions, common to both code
   --           and annotations
   --
   -- annotation_named_argument_association
   -- named_argument_association
   -- annotation_positional_argument_association
   -- positional_argument_association
   -- annotation_attribute_designator_opt
   -- attribute_designator_opt
   -- annotation_aggregate_or_expression
   -- aggregate_or_expression
   -- annotation_primary
   -- primary
   -- annotation_qualified_expression
   -- qualified_expression
   --
   -- Group 3 - Expressions appearing in annotations only
   --
   -- store
   -- store_list
   -- return_expression
   -- predicate
   ------------------------------------------------------------------------

   case Syntax_Node_Type (Node => Parent_Node (Current_Node => Exp_Node)) is

      when SPSymbols.constant_declaration |
        SPSymbols.variable_declaration |
        SPSymbols.assignment_statement |
        SPSymbols.condition            |
        SPSymbols.case_statement       |
        SPSymbols.return_statement     |
        SPSymbols.delay_statement      =>

         -- Context here is passed in from calling environment, and
         -- no change required.
         New_Context_Type := TypeContextStack.Top (T_Stack);

      when SPSymbols.annotation_named_argument_association | SPSymbols.named_argument_association =>

         if ExpStack.IsEmpty (E_Stack) then
            -- Must be a named argument association in a procedure
            -- call.  Wf_Proc_Call processes the parameter name, and
            -- does NOT put it on the expression stack, but it does pass
            -- the expected type into WalkExpression, so this is unchanged
            -- in this case.
            New_Context_Type := TypeContextStack.Top (T_Stack);
         else
            -- Must be a named argument association which is the parameter
            -- list of a function call.
            TopOfExpStack := ExpStack.Top (E_Stack);

            -- We need to find the parameter name, and determine if it is legal:
            FindNamedArgumentAssociationParameter
              (Node                => Parent_Node (Current_Node => Exp_Node),
               SubprogSym          => TopOfExpStack.Other_Symbol,
               IsAnnotation        => (Syntax_Node_Type (Node => Parent_Node (Current_Node => Exp_Node)) =
                    SPSymbols.annotation_named_argument_association),
               NameIsParameterName => ParameterNameOK,
               ParamSym            => ParamSymbol);

            if ParameterNameOK then
               -- The parameter name denotes a legal parameter of this subprogram,
               -- so look up its type.
               New_Context_Type := Dictionary.GetType (ParamSymbol);

            else
               -- The parameter name is illegal.  This will be picked up again later
               -- on in wf_named_argument_association.  The type context simply
               -- becomes unknown.
               New_Context_Type := Dictionary.GetUnknownTypeMark;
            end if;
         end if;

      when SPSymbols.qualified_expression | SPSymbols.annotation_qualified_expression =>

         -- Context changes to the type denoted by the Name preceeding the
         -- expression.  The result of evaluating this Name should be on the
         -- top of the Expression stack.
         TopOfExpStack    := ExpStack.Top (E_Stack);
         New_Context_Type := TopOfExpStack.Type_Symbol;

      when SPSymbols.positional_argument_association | SPSymbols.annotation_positional_argument_association =>

         -- Parent could be name_argument_list or another
         -- positional_argument_association.
         -- This could be part of a
         --   Type conversion
         --   Array index
         --   Function call

         if ExpStack.IsEmpty (E_Stack) then
            -- EStack might be empty here => we must be processing an
            -- actual param of a procedure call statement.  Context
            -- will have been passed in from wf_proc_call, so no change here.
            New_Context_Type := TypeContextStack.Top (T_Stack);
         else
            TopOfExpStack := ExpStack.Top (E_Stack);

            case TopOfExpStack.Sort is
               when Is_Object =>
                  -- Must be an array indexing expression.  The context is therefore
                  -- the subtype corresponding to the N'th dimension of that array type.
                  New_Context_Type := Dictionary.GetArrayIndex (TopOfExpStack.Type_Symbol, TopOfExpStack.Param_Count + 1);

               when Is_Function =>
                  -- Must be an actual parameter of a function call.  The context
                  -- is the subtype indicated by the corresponding formal parameter,
                  -- if there is any such parameter.  If the wrong number of actual
                  -- parameters has been given, then return UnknownTypeMark - this
                  -- error will be picked up later in the UP pass.

                  if TopOfExpStack.Param_Count < Dictionary.GetNumberOfSubprogramParameters (TopOfExpStack.Other_Symbol) then

                     New_Context_Type :=
                       Dictionary.GetType
                       (Dictionary.GetSubprogramParameter (TopOfExpStack.Other_Symbol, TopOfExpStack.Param_Count + 1));
                  else
                     New_Context_Type := Dictionary.GetUnknownTypeMark;
                  end if;

               when Is_Type_Mark =>
                  -- Must be a type conversion.  The argument of the type conversion
                  -- could be pretty much anything - legal or illegal, so the
                  -- context is unknown.
                  New_Context_Type := Dictionary.GetUnknownTypeMark;

               when others =>
                  -- Other cases - all errors which will be caught later on...
                  -- We still need to push something onto the TStack to make
                  -- it balance, so simply copy the existing top entry.
                  New_Context_Type := TypeContextStack.Top (T_Stack);
            end case;
         end if;

      when SPSymbols.aggregate_or_expression | SPSymbols.annotation_aggregate_or_expression =>

         case Syntax_Node_Type (Node => Parent_Node (Current_Node => Parent_Node (Current_Node => Exp_Node))) is
            when SPSymbols.component_association | SPSymbols.annotation_component_association =>
               -- Must be an array aggregate with a single others clause, so the
               -- new context type is the type of the array element
               TopOfExpStack    := ExpStack.Top (E_Stack);
               New_Context_Type := Dictionary.GetArrayComponent (TopOfExpStack.Type_Symbol);

            when SPSymbols.named_association                |
              SPSymbols.annotation_named_association     |
              SPSymbols.named_association_rep            |
              SPSymbols.annotation_named_association_rep =>

               TopOfExpStack := ExpStack.Top (E_Stack);

               case TopOfExpStack.Sort is
                  when Is_Parameter_Name =>
                     -- Must be a named field of a record aggregate.  New context is
                     -- the type of that field, unless the field was itself illegal,
                     -- in which case the context is unknown.
                     if TopOfExpStack = NullParameterRecord then
                        New_Context_Type := Dictionary.GetUnknownTypeMark;
                     else
                        New_Context_Type := Dictionary.GetType (TopOfExpStack.Other_Symbol);
                     end if;

                  when Is_Type_Mark =>
                     -- Must be a named element of an array aggregate.  New context is
                     -- the type of the array element.
                     New_Context_Type := Dictionary.GetArrayComponent (TopOfExpStack.Type_Symbol);

                  when others =>
                     New_Context_Type := TypeContextStack.Top (T_Stack);
               end case;

            when SPSymbols.positional_association                |
              SPSymbols.positional_association_rep            |
              SPSymbols.annotation_positional_association     |
              SPSymbols.annotation_positional_association_rep =>

               TopOfExpStack := ExpStack.Top (E_Stack);

               case TopOfExpStack.Sort is
                  when Is_Type_Mark =>

                     if Dictionary.TypeIsRecord (TopOfExpStack.Type_Symbol) then
                        -- New context is the type of the N'th field of the record, assuming
                        -- there is such a field.
                        if (TopOfExpStack.Param_Count + 1) <= Dictionary.GetNumberOfComponents (TopOfExpStack.Type_Symbol) then

                           New_Context_Type :=
                             Dictionary.GetType
                             (Dictionary.GetRecordComponent (TopOfExpStack.Type_Symbol, TopOfExpStack.Param_Count + 1));
                        else
                           New_Context_Type := Dictionary.GetUnknownTypeMark;
                        end if;

                     elsif Dictionary.TypeIsArray (TopOfExpStack.Type_Symbol) then
                        -- New context is the element type of the array
                        New_Context_Type := Dictionary.GetArrayComponent (TopOfExpStack.Type_Symbol);
                     else
                        -- Must be an error - this will be caught later on in the UP
                        -- pass, but we need to push something so...
                        New_Context_Type := TypeContextStack.Top (T_Stack);
                     end if;

                  when others =>
                     -- Must be an error - this will be caught later on in the UP
                     -- pass, but we need to push something so...
                     New_Context_Type := TypeContextStack.Top (T_Stack);
               end case;

            when others =>
               New_Context_Type := TypeContextStack.Top (T_Stack);
         end case;

      when SPSymbols.annotation_attribute_designator_opt | SPSymbols.attribute_designator_opt =>

         -- Context change for attribute arguments is handled in
         -- AttributeDesignatorTypeFromContext, so no change here.
         New_Context_Type := TypeContextStack.Top (T_Stack);

      when SPSymbols.annotation_primary | SPSymbols.primary =>

         -- Must be a parenthesized expression - context does not change.
         New_Context_Type := TypeContextStack.Top (T_Stack);

      when SPSymbols.store =>

         TopOfExpStack := ExpStack.Top (E_Stack);

         case TopOfExpStack.Sort is

            when Is_Object =>

               if Dictionary.TypeIsArray (TopOfExpStack.Type_Symbol) then

                  New_Context_Type := Dictionary.GetArrayComponent (TopOfExpStack.Type_Symbol);

               elsif Dictionary.TypeIsRecord (TopOfExpStack.Type_Symbol) then

                  -- The record field being updated should be an identifier
                  -- node directly below the store node, so
                  TNode2 := Last_Child_Of (Start_Node => Parent_Node (Current_Node => Exp_Node));

                  if Syntax_Node_Type (Node => TNode2) = SPSymbols.identifier then
                     ParamSymbol :=
                       Dictionary.LookupSelectedItem
                       (Prefix   => TopOfExpStack.Type_Symbol,
                        Selector => Node_Lex_String (Node => TNode2),
                        Scope    => Dictionary.GetScope (TopOfExpStack.Type_Symbol),
                        Context  => Dictionary.ProofContext);

                     if ParamSymbol /= Dictionary.NullSymbol and then Dictionary.IsRecordComponent (ParamSymbol) then
                        STree.Set_Node_Lex_String (Sym  => ParamSymbol,
                                                   Node => TNode2);
                        New_Context_Type := Dictionary.GetType (ParamSymbol);
                     else
                        -- error - will be caught in up_wf_store
                        New_Context_Type := TypeContextStack.Top (T_Stack);
                     end if;
                  else
                     -- error - will be caught in up_wf_store
                     New_Context_Type := TypeContextStack.Top (T_Stack);
                  end if;
               else
                  -- error - will be caught in up_wf_store
                  New_Context_Type := TypeContextStack.Top (T_Stack);
               end if;

            when others =>
               -- Must be an error - this will be caught later on in the UP
               -- pass, but we need to push something so...
               New_Context_Type := TypeContextStack.Top (T_Stack);
         end case;

      when SPSymbols.store_list =>

         TopOfExpStack := ExpStack.Top (E_Stack);

         case TopOfExpStack.Sort is

            when Is_Object =>

               if Dictionary.TypeIsArray (TopOfExpStack.Type_Symbol) then

                  New_Context_Type := Dictionary.GetArrayIndex (TopOfExpStack.Type_Symbol, TopOfExpStack.Param_Count + 1);

               else
                  -- error - will be caught in up_wf_store
                  New_Context_Type := TypeContextStack.Top (T_Stack);
               end if;

            when others =>
               -- Must be an error - this will be caught later on in the UP
               -- pass, but we need to push something so...
               New_Context_Type := TypeContextStack.Top (T_Stack);
         end case;

      when SPSymbols.return_expression | SPSymbols.predicate =>
         -- Context for predicate and return_expression is always passed
         -- in from wf_predicate, or wf_function_constraint, so no change needed.
         New_Context_Type := TypeContextStack.Top (T_Stack);

      when others =>
         -- In all other cases, the context is unchanged, but we push a copy
         -- of the current context type to keep the stack balanced.
         New_Context_Type := TypeContextStack.Top (T_Stack);
   end case;

end ExpressionTypeFromContext;
